<?php
/*
Dragon Go Server
Copyright (C) 2001-2014  Erik Ouchterlony, Rod Ival

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

$TranslateGroups[] = "Game";


/*
Basic (EBNF) Definition

  "..." : terminal symbols
  [...] : option: occurs at most once
  {...} : repetition: any number of times, including zero
  (...) : grouping
    |   : exclusive or
 italics: parameter explained at some other place


SGF EBNF Definition

  Collection = GameTree { GameTree }
  GameTree   = "(" Sequence { GameTree } ")"
  Sequence   = Node { Node }
  Node       = ";" { Property }
  Property   = PropIdent PropValue { PropValue }
  PropIdent  = UcLetter { UcLetter }
  PropValue  = "[" CValueType "]"
  CValueType = (ValueType | Compose)
  ValueType  = (None | Number | Real | Double | Color | SimpleText |
      Text | Point  | Move | Stone)


White space (space, tab, carriage return, line feed, vertical tab and so on)
 may appear anywhere between PropValues, Properties, Nodes, Sequences and GameTrees.
There are two types of property lists: 'list of' and 'elist of'.
'list of':    PropValue { PropValue }
'elist of':   ((PropValue { PropValue }) | None)
              In other words elist is list or "[]".


Property Value Types

  UcLetter   = "A".."Z"
  Digit      = "0".."9"
  None       = ""

  Number     = [("+"|"-")] Digit { Digit }
  Real       = Number ["." Digit { Digit }]

  Double     = ("1" | "2")
  Color      = ("B" | "W")

  SimpleText = { any character (handling see below) }
  Text       = { any character (handling see below) }

  Point      = game-specific
  Move       = game-specific
  Stone      = game-specific

  Compose    = ValueType ":" ValueType


*/

define('SGF_VAR_BEG', '(');
define('SGF_VAR_END', ')');
define('SGF_ARG_BEG', '[');
define('SGF_ARG_END', ']');
define('SGF_NOD_BEG', ';');
define('SGF_VAR_KEY', '++');


// Explode a sgf string ($sgf, from file) into an array of games ($games)
function sgf_parser( $sgf, &$games)
{
   $err= '';
   $games= array();

   $sgf_len= strlen($sgf);
   $i= 0;
   while ( !$err && false !== ($i= strpos($sgf,SGF_VAR_BEG,$i)) )
   {
      $ivar= 0;
      $vars[$ivar]= array();

      $i++;
      if ( !sgf_skip_space( $sgf, $i, $sgf_len) )
      {
         $err = T_('Bad end of file#sgf');
         break;
      }

      if ( $sgf{$i} != SGF_NOD_BEG )
      {
         $err = T_('Bad root node#sgf');
         break;
      }

      $i++;
      $err = sgf_parse_node( $sgf, $i, $sgf_len, $node);
      if ( $err )
         break;

      if ( !isset($node['GM']) || @$node['GM'][0] != 1 )
      {
         $err = T_('Not a Go game (GM[1])#sgf');
         break;
      }

      foreach ( array( //root default values
           'FF' => 4,
           'GM' => 1,
           'SZ' => 19,
           'RU' => 'Japanese',
         ) as $key => $arg)
      {
         if ( !isset($node[$key]) )
            $node[$key]= array($arg);
      }

      $vars[$node_ivar=$ivar][]= $node;

      while ( !$err && $ivar>=0 && $c=sgf_skip_space( $sgf, $i, $sgf_len) )
      {
         switch ( (string)$c )
         {
            case SGF_VAR_END:
               $ivar--; $i++;
               $vars[$ivar][SGF_VAR_KEY][]= $vars[$ivar+1];
               break;

            case SGF_VAR_BEG:
               $i++;
               if ( SGF_NOD_BEG != sgf_skip_space( $sgf, $i, $sgf_len) )
               {
                  $err = T_('Bad node start#sgf');
                  break;
               }
               if ( $node_ivar <= $ivar )
                  $vars[$ivar][SGF_VAR_KEY]= array();
               $ivar++;
               $vars[$ivar]= array();
            case SGF_NOD_BEG :
               $i++;
               $err= sgf_parse_node( $sgf, $i, $sgf_len, $node);
               $vars[$node_ivar=$ivar][]= $node;
               break;

            default :
               $err = T_('Syntax error#sgf');
               break;
         }
      }
      if ( $err )
         break;

      if ( $ivar>=0 )
      {
         $err = T_('Missing right parenthesis#sgf');
         break;
      }

      $games[]= $vars[0];
   }

   if ( $err )
   {
      $tmp= max(0,$i-20);
      $err= "SGF: ".$err
         . "\nnear: "
         . ltrim(substr( $sgf, $tmp, $i-$tmp))
         . '<'
         . substr( $sgf, $i, 1)
         . '>'
         . rtrim(substr( $sgf, $i+1, 20))
         ;
   }
   return $err;
}//sgf_parser


// In short, it does the opposite of sgf_parser()
function sgf_builder( $games, &$sgf)
{
   $err= '';
   $sgf= '';

   $vars= array();
   //$games is an array of games (i.e. variations)
   for ( $i=count($games)-1; $i>=0; $i-- )
   {
      sgf_var_push( $vars, $games[$i]);
   }

   while ( $var=array_pop($vars) )
   {
      if ( $var === SGF_VAR_END ) //see sgf_var_push()
      {
         $sgf.= SGF_VAR_END."\r\n";
         continue;
      }

      $sgf.= SGF_VAR_BEG."\r\n";
      //a variation is an array of nodes
      foreach ( $var as $id => $node )
      {
         if ( $id === SGF_VAR_KEY )
         {
            //this perticular node is an array of variations
            for ( $i=count($node)-1; $i>=0; $i-- )
            {
               sgf_var_push( $vars, $node[$i]);
            }
            continue;
         }

         $sgf.= SGF_NOD_BEG."\r\n";
         //a node is an array of properties
         foreach ( $node as $key => $args )
         {
            $sgf.= "$key";
            foreach ( $args as $arg )
            {
               $sgf.= SGF_ARG_BEG.$arg.SGF_ARG_END;
            }
            $sgf.= "\r\n";
         }
      }
   }

   return $err;
}


function sgf_var_push( &$vars, &$var)
{
   if ( is_array($var) )
   {
      $vars[]= SGF_VAR_END;
      $vars[]= &$var;
   }
}


function sgf_skip_space( $sgf, &$i, $l)
{
   while ( $i<$l )
   {
      if ( $sgf{$i} > ' ' )
         return $sgf{$i};
      $i++;
   }
   return 0;
}


function sgf_parse_node( $sgf, &$i, $l, &$node)
{
   $node= array();
   while ( $key= sgf_parse_key( $sgf, $i, $l) )
   {
      $err= sgf_parse_args( $sgf, $i, $l, $args, $key);
      if ( $err )
         return $err;
      $node[$key]= $args;
   }
   $c= $sgf{$i}; //sgf_skip_space( $sgf, $i, $l)
   if ( $c != SGF_NOD_BEG &&  $c != SGF_VAR_BEG && $c != SGF_VAR_END )
   {
      return T_('Syntax error#sgf');
   }
   return '';
}


function sgf_parse_args( $sgf, &$i, $l, &$args, $key)
{
   $args= array();
   while ( sgf_skip_space( $sgf, $i, $l) == SGF_ARG_BEG )
   {
      $j= $i;
      while ( false !== ($j=strpos($sgf,SGF_ARG_END,$j+1)) )
      {
         if ( $sgf{$j-1} != '\\' )
         {
            $arg= substr($sgf, $i+1, $j-$i-1);
            $args[]= $arg;
            $i= $j+1;
            break;
         }
      }
      if ( false === $j )
         return T_('Missing right bracket#sgf');
   }
   return '';
}


function sgf_parse_key( $sgf, &$i, $l)
{
   $key= '';
   while ( $i<$l )
   {
      $c= $sgf{$i};
      if ( $c>='A' && $c<='Z' )
         $key.= $c;
      else if ( ($c>' ') && ($c<'a' || $c>'z') )
         break;
      $i++;
   }
   return $key;
}


if ( defined('ENABLE_STDHANDICAP') && ENABLE_STDHANDICAP ) {

function handicap_push( &$vars, &$var, $nb)
{
   if ( is_array($var) )
   {
      $vars[]= array($nb, &$var);
   }
}

// Read the standard handicap pattern file
// and convert it to a stonestring.
function get_handicap_pattern( $size, $handicap, &$err)
{
   $stonestring ='';
   if ( $handicap < 2 )
      return $stonestring;

   $game = array();

   $filename = "pattern/standard_handicap_$size.sgf";
   $sgf = @read_from_file( $filename, 0);
   if ( is_string($sgf) )
      $err = sgf_parser( $sgf, $game);
   else
      $err = T_('File not found');
   if ( $err ) {
      //Simply returning the error message will allow
      //the player to manually add his handicap stones.
      $err= sprintf( T_('Bad handicap pattern for %s'), "size=$size h=$handicap err=[$err]" );
      return $stonestring;
   }
   $game= $game[0]; //keep the first game only

   $nb = 0;
   $vars = array();
   handicap_push( $vars, $game, $nb);

   while ( list($nb,$var)=array_pop($vars) )
   {
      $stonestring = substr( $stonestring, 0, 2*$nb);

      //a variation is an array of nodes
      foreach ( $var as $id => $node )
      {
         if ( $id === SGF_VAR_KEY )
         {
            //this perticular node is an array of variations
            for ( $i=count($node)-1; $i>=0; $i-- )
               handicap_push( $vars, $node[$i], $nb);
            continue;
         }

         //a node is an array of properties
         if ( isset($node['B']) || isset($node['W']) )
         {
            $co= @$node['B'][0];
            if ( !$co )
               $co= @$node['W'][0];
            if ( strlen($co) != 2 )
               return $stonestring;
            $stonestring.= $co;
            $nb++;
            if ( $nb >= $handicap )
               return $stonestring;
         }
      }
   }
   //See previous error comment
   $err = sprintf( T_('Insufficient handicap pattern for %s'), "size=$size h=$handicap n=$nb" );
   return $stonestring;
}

}//ENABLE_STDHANDICAP


/*! \brief Contains parsed properties from SGF for further processing. */
class SgfParser
{
   public $error; // error-message | '' (=success)
   public $Size = 0;
   public $Handicap = 0; // number of handicap-stones
   public $Komi = 0;
   public $SetWhite = array(); // [ sgf-coord, ... ]
   public $SetBlack = array();
   public $Moves = array(); // [ 'B|W'.sgf-coord, ... ], for example: [ 'Baa', ... ]

   public function __construct( $error )
   {
      $this->error = $error;
   }


   // ------------ static functions ----------------------------

   /*! Returns false if passed SGF-data does not rudimentary looking like a SGF. */
   public static function might_be_sgf( $sgf_data )
   {
      return preg_match("/^\\s*\\(\\s*;\\s*[a-z]/si", $sgf_data);
   }

   /*!
    * \brief Parses SGF-data into resulting-array (used to load SGF and flatten into Goban-objects for Shape-game).
    * \return SgfParser-instance with filled properties; parsing-error in SgfParser->Error or '' if ok
    */
   public static function parse_sgf( $sgf_data )
   {
      $game = array();
      $error = sgf_parser( $sgf_data, $game );

      $sgf_parser = new SgfParser( $error );
      if ( $error )
         return $sgf_parser;

      $game = $game[0]; // check 1st game only
      $movenum = 0; // current move-number
      $vars = array(); // variations
      handicap_push( $vars, $game, $movenum );

      $parsed_HA = $parsed_KM = null;
      while ( list($movenum, $var) = array_pop($vars) ) // process variations-stack
      {
         // a variation is an array of nodes
         foreach ( $var as $id => $node )
         {
            if ( $id === SGF_VAR_KEY )
            {
               // this particular node is an array of variations, but only take first var (main-branch)
               handicap_push( $vars, $node[0], $movenum );
               continue;
            }

            // a node is an array of properties
            if ( isset($node['B']) || isset($node['W']) )
            {
               $key = ( isset($node['B']) ) ? 'B' : 'W';
               $sgf_coord = @$node[$key][0];
               $sgf_parser->Moves[] = $key . $sgf_coord;
               $movenum++;
            }
            if ( isset($node['AB']) )
            {
               foreach ( @$node['AB'] as $sgf_coord )
                  $sgf_parser->SetBlack[] = $sgf_coord;
            }
            if ( isset($node['AW']) )
            {
               foreach ( @$node['AW'] as $sgf_coord )
                  $sgf_parser->SetWhite[] = $sgf_coord;
            }
            if ( isset($node['SZ']) && !$sgf_parser->Size )
               $sgf_parser->Size = (int)$node['SZ'][0];
            if ( isset($node['HA']) && is_null($parsed_HA) )
               $sgf_parser->Handicap = $parsed_HA = (int)$node['HA'][0];
            if ( isset($node['KM']) && is_null($parsed_KM) )
               $sgf_parser->Komi = $parsed_KM = (float)$node['KM'][0];
         }
      }

      return $sgf_parser;
   }//parse_sgf

}//end 'SgfParser'

?>
